@@ -10,31 +10,36 @@
 
 
 #include "stdafx.h"
-#ifdef MODPLUG_TRACKER
-#include "../mptrack/Mptrack.h"	// For CTrackApp::OpenURL
-#include "../mptrack/TrackerSettings.h"
-#include "../mptrack/Moddoc.h"
-#include "../mptrack/Reporting.h"
-#include "../mptrack/Mainfrm.h"
-#endif // MODPLUG_TRACKER
-#ifdef MPT_EXTERNAL_SAMPLES
-#include "../common/mptFileIO.h"
-#endif // MPT_EXTERNAL_SAMPLES
-#include "../common/version.h"
-#include "../soundlib/AudioCriticalSection.h"
-#include "../common/serialization_utils.h"
 #include "Sndfile.h"
-#include "Tables.h"
+#include "Container.h"
 #include "mod_specifications.h"
+#include "OPL.h"
+#include "Tables.h"
 #include "tuningcollection.h"
 #include "plugins/PluginManager.h"
 #include "plugins/PlugInterface.h"
-#include "../common/mptStringBuffer.h"
 #include "../common/FileReader.h"
-#include "Container.h"
-#include "OPL.h"
+#include "../common/mptStringBuffer.h"
+#include "../common/serialization_utils.h"
+#include "../common/version.h"
+#include "../soundlib/AudioCriticalSection.h"
 #include "mpt/io/io.hpp"
 #include "mpt/io/io_stdstream.hpp"
+#include "mpt/random/seed.hpp"
+
+#ifdef MODPLUG_TRACKER
+#include "../mptrack/Mainfrm.h"
+#include "../mptrack/Moddoc.h"
+#include "../mptrack/Mptrack.h"	// For CTrackApp::OpenURL
+#include "../mptrack/Reporting.h"
+#include "../mptrack/TrackerSettings.h"
+#endif // MODPLUG_TRACKER
+
+#ifdef MPT_EXTERNAL_SAMPLES
+#include "../common/mptFileIO.h"
+#include "mpt/io_file/inputfile.hpp"
+#include "mpt/io_file_read/inputfile_filecursor.hpp"
+#endif // MPT_EXTERNAL_SAMPLES
 
 #ifndef NO_ARCHIVE_SUPPORT
 #include "../unarchiver/unarchiver.h"
@@ -54,26 +59,47 @@ bool SettingCacheCompleteFileBeforeLoading()
 }
 
 
-mpt::ustring FileHistory::AsISO8601() const
+mpt::ustring FileHistory::AsISO8601(mpt::Date::LogicalTimezone internalTimezone) const
 {
-	tm date = loadDate;
 	if(openTime > 0)
 	{
 		// Calculate the date when editing finished.
 		double openSeconds = static_cast<double>(openTime) / HISTORY_TIMER_PRECISION;
-		tm tmpLoadDate = loadDate;
-		int64 loadDateSinceEpoch = mpt::Date::Unix::FromUTC(tmpLoadDate);
-		int64 saveDateSinceEpoch = loadDateSinceEpoch + mpt::saturate_round<int64>(openSeconds);
-		date = mpt::Date::Unix(saveDateSinceEpoch).AsUTC();
+		mpt::Date::AnyGregorian tmpLoadDate = loadDate;
+		if (internalTimezone == mpt::Date::LogicalTimezone::UTC)
+		{
+			int64 loadDateSinceEpoch = mpt::Date::UnixAsSeconds(mpt::Date::UnixFromUTC(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::UTC>(tmpLoadDate)));
+			int64 saveDateSinceEpoch = loadDateSinceEpoch + mpt::saturate_round<int64>(openSeconds);
+			return mpt::Date::ToShortenedISO8601(mpt::Date::UnixAsUTC(mpt::Date::UnixFromSeconds(saveDateSinceEpoch)));
+#ifdef MODPLUG_TRACKER
+		} else if(internalTimezone == mpt::Date::LogicalTimezone::Local)
+		{
+			int64 loadDateSinceEpoch = mpt::Date::UnixAsSeconds(mpt::Date::UnixFromLocal(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::Local>(tmpLoadDate)));
+			int64 saveDateSinceEpoch = loadDateSinceEpoch + mpt::saturate_round<int64>(openSeconds);
+			return mpt::Date::ToShortenedISO8601(mpt::Date::UnixAsLocal(mpt::Date::UnixFromSeconds(saveDateSinceEpoch)));
+#endif // MODPLUG_TRACKER
+		} else
+		{
+			// assume UTC for unspecified timezone when calculating
+			int64 loadDateSinceEpoch = mpt::Date::UnixAsSeconds(mpt::Date::UnixFromUTC(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::UTC>(tmpLoadDate)));
+			int64 saveDateSinceEpoch = loadDateSinceEpoch + mpt::saturate_round<int64>(openSeconds);
+			return mpt::Date::ToShortenedISO8601(mpt::Date::forget_timezone(mpt::Date::UnixAsUTC(mpt::Date::UnixFromSeconds(saveDateSinceEpoch))));
+		}
+	} else
+	{
+		if(internalTimezone == mpt::Date::LogicalTimezone::UTC)
+		{
+			return mpt::Date::ToShortenedISO8601(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::UTC>(loadDate));
+#ifdef MODPLUG_TRACKER
+		} else if(internalTimezone == mpt::Date::LogicalTimezone::Local)
+		{
+			return mpt::Date::ToShortenedISO8601(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::Local>(loadDate));
+#endif // MODPLUG_TRACKER
+		} else
+		{
+			return mpt::Date::ToShortenedISO8601(loadDate);
+		}
 	}
-	return mpt::Date::ToShortenedISO8601(date);
-}
-
-
-CSoundFile::PlayState::PlayState()
-{
-	std::fill(std::begin(Chn), std::end(Chn), ModChannel{});
-	m_midiMacroScratchSpace.reserve(kMacroLength);  // Note: If macros ever become variable-length, the scratch space needs to be at least one byte longer than the longest macro in the file for end-of-SysEx insertion to stay allocation-free in the mixer!
 }
 
 
@@ -91,12 +117,12 @@ CSoundFile::CSoundFile() :
 	m_pModSpecs(&ModSpecs::itEx),
 	m_nType(MOD_TYPE_NONE),
 	Patterns(*this),
-#ifdef MODPLUG_TRACKER
-	m_MIDIMapper(*this),
-#endif
 	Order(*this),
 	m_PRNG(mpt::make_prng<mpt::fast_prng>(mpt::global_prng())),
 	m_visitedRows(*this)
+#ifdef MODPLUG_TRACKER
+	, m_MIDIMapper(*this)
+#endif
 {
 	MemsetZero(MixSoundBuffer);
 	MemsetZero(MixRearBuffer);
@@ -144,14 +170,21 @@ void CSoundFile::AddToLog(LogLevel level, const mpt::ustring &text) const
 
 
 // Global variable initializer for loader functions
-void CSoundFile::InitializeGlobals(MODTYPE type)
+void CSoundFile::InitializeGlobals(MODTYPE type, CHANNELINDEX numChannels)
 {
 	// Do not add or change any of these values! And if you do, review each and every loader to check if they require these defaults!
 	m_nType = type;
+	MPT_ASSERT(numChannels <= MAX_BASECHANNELS);
+	LimitMax(numChannels, MAX_BASECHANNELS);
 
 	MODTYPE bestType = GetBestSaveFormat();
 	m_playBehaviour = GetDefaultPlaybackBehaviour(bestType);
-	SetModSpecsPointer(m_pModSpecs, bestType);
+	if(bestType == MOD_TYPE_IT && type != bestType)
+	{
+		// This is such an odd behaviour that it's unlikely that any of the other formats will need it by default. Re-enable as needed.
+		m_playBehaviour.reset(kITInitialNoteMemory);
+	}
+	m_pModSpecs = &GetModSpecifications(bestType);
 
 	// Delete instruments in case some previously called loader already created them.
 	for(INSTRUMENTINDEX i = 1; i <= m_nInstruments; i++)
@@ -161,14 +194,11 @@ void CSoundFile::InitializeGlobals(MODTYPE type)
 	}
 
 	m_ContainerType = ModContainerType::None;
-	m_nChannels = 0;
 	m_nInstruments = 0;
 	m_nSamples = 0;
 	m_nSamplePreAmp = 48;
 	m_nVSTiVolume = 48;
 	m_OPLVolumeFactor = m_OPLVolumeFactorScale;
-	m_nDefaultSpeed = 6;
-	m_nDefaultTempo.Set(125);
 	m_nDefaultGlobalVolume = MAX_GLOBAL_VOLUME;
 	m_SongFlags.reset();
 	m_nMinPeriod = 16;
@@ -176,12 +206,14 @@ void CSoundFile::InitializeGlobals(MODTYPE type)
 	m_nResampling = SRCMODE_DEFAULT;
 	m_dwLastSavedWithVersion = Version(0);
 	m_dwCreatedWithVersion = Version(0);
+	m_nTempoMode = TempoMode::Classic;
 
 	SetMixLevels(MixLevels::Compatible);
 
-	Patterns.ClearPatterns();
+	Patterns.DestroyPatterns();
 	Order.Initialize();
 
+	m_globalScript.clear();
 	m_songName.clear();
 	m_songArtist.clear();
 	m_songMessage.clear();
@@ -194,18 +226,11 @@ void CSoundFile::InitializeGlobals(MODTYPE type)
 
 	// Note: we do not use the Amiga resampler for DBM as it's a multichannel format and can make use of higher-quality Amiga soundcards instead of Paula.
 	if(GetType() & (/*MOD_TYPE_DBM | */MOD_TYPE_DIGI | MOD_TYPE_MED | MOD_TYPE_MOD | MOD_TYPE_OKT | MOD_TYPE_SFX | MOD_TYPE_STP))
-	{
 		m_SongFlags.set(SONG_ISAMIGA);
-	}
-}
+	if(GetType() & (MOD_TYPE_AMF0 | MOD_TYPE_DIGI | MOD_TYPE_MTM))
+		m_SongFlags.set(SONG_FORMAT_NO_VOLCOL);
 
-
-void CSoundFile::InitializeChannels()
-{
-	for(CHANNELINDEX nChn = 0; nChn < MAX_BASECHANNELS; nChn++)
-	{
-		InitChannel(nChn);
-	}
+	ChnSettings.assign(numChannels, {});
 }
 
 
@@ -215,7 +240,7 @@ struct FileFormatLoader
 	decltype(&CSoundFile::ReadXM) loader;
 };
 
-#ifdef MODPLUG_TRACKER
+#if defined(MODPLUG_TRACKER) && !defined(MPT_BUILD_DEBUG)
 #define MPT_DECLARE_FORMAT(format) { nullptr, &CSoundFile::Read ## format }
 #else
 #define MPT_DECLARE_FORMAT(format) { CSoundFile::ProbeFileHeader ## format, &CSoundFile::Read ## format }
@@ -263,6 +288,8 @@ static constexpr FileFormatLoader ModuleFormatLoaders[] =
 	MPT_DECLARE_FORMAT(PLM),
 	MPT_DECLARE_FORMAT(AM),
 	MPT_DECLARE_FORMAT(J2B),
+	MPT_DECLARE_FORMAT(GT2),
+	MPT_DECLARE_FORMAT(GTK),
 	MPT_DECLARE_FORMAT(PT36),
 	MPT_DECLARE_FORMAT(SymMOD),
 	MPT_DECLARE_FORMAT(MUS_KM),
@@ -271,12 +298,26 @@ static constexpr FileFormatLoader ModuleFormatLoaders[] =
 	MPT_DECLARE_FORMAT(STP),
 	MPT_DECLARE_FORMAT(DSym),
 	MPT_DECLARE_FORMAT(STX),
+	MPT_DECLARE_FORMAT(UNIC),  // Magic bytes clash with MOD, must be tried first
 	MPT_DECLARE_FORMAT(MOD),
 	MPT_DECLARE_FORMAT(ICE),
+	MPT_DECLARE_FORMAT(KRIS),
 	MPT_DECLARE_FORMAT(669),
+	MPT_DECLARE_FORMAT(667),
 	MPT_DECLARE_FORMAT(C67),
 	MPT_DECLARE_FORMAT(MO3),
-	MPT_DECLARE_FORMAT(M15),
+	MPT_DECLARE_FORMAT(FC),
+	MPT_DECLARE_FORMAT(FTM),
+	MPT_DECLARE_FORMAT(RTM),
+	MPT_DECLARE_FORMAT(TCB),
+	MPT_DECLARE_FORMAT(CBA),
+	MPT_DECLARE_FORMAT(ETX),
+	MPT_DECLARE_FORMAT(DSm),
+	MPT_DECLARE_FORMAT(STK),
+	MPT_DECLARE_FORMAT(XMF),
+	MPT_DECLARE_FORMAT(Puma),
+	MPT_DECLARE_FORMAT(GMC),
+	MPT_DECLARE_FORMAT(IMS),
 };
 
 #undef MPT_DECLARE_FORMAT
@@ -365,7 +406,7 @@ CSoundFile::ProbeResult CSoundFile::Probe(ProbeFlags flags, mpt::span<const std:
 	{
 		if((result == ProbeWantMoreData) && (data.size() >= ProbeRecommendedSize))
 		{
-			// If the prober wants more daat but we already provided the recommended required maximum,
+			// If the prober wants more data but we already provided the recommended required maximum,
 			// just return success as this is the best we can do for the suggestesd probing size.
 			result = ProbeSuccess;
 		}
@@ -374,6 +415,14 @@ CSoundFile::ProbeResult CSoundFile::Probe(ProbeFlags flags, mpt::span<const std:
 }
 
 
+void CSoundFile::Create(MODTYPE type, CHANNELINDEX numChannels, CModDoc *modDoc)
+{
+	Create(FileReader{}, CSoundFile::loadCompleteModule, modDoc);
+	SetType(type);
+	ChnSettings.resize(numChannels);
+}
+
+
 bool CSoundFile::Create(FileReader file, ModLoadingFlags loadFlags, CModDoc *pModDoc)
 {
 	m_nMixChannels = 0;
@@ -398,7 +447,18 @@ bool CSoundFile::Create(FileReader file, ModLoadingFlags loadFlags, CModDoc *pMo
 		CUnarchiver unarchiver(file);
 		if(unarchiver.ExtractBestFile(GetSupportedExtensions(true)))
 		{
-			if(CreateInternal(unarchiver.GetOutputFile(), loadFlags))
+			FileReader outputFile = unarchiver.GetOutputFile();
+#if defined(MPT_WITH_ANCIENT)
+			// Special case for MMCMP/PP20/XPK/etc. inside ZIP/RAR/LHA
+			std::unique_ptr<CAncientArchive> ancientArchive;
+			if(!unarchiver.IsKind<CAncientArchive>())
+			{
+				ancientArchive = std::make_unique<CAncientArchive>(outputFile);
+				if(ancientArchive->IsArchive() && ancientArchive->ExtractFile(0))
+					outputFile = ancientArchive->GetOutputFile();
+			}
+#endif
+			if(CreateInternal(outputFile, loadFlags))
 			{
 				// Read archive comment if there is no song comment
 				if(m_songMessage.empty())
@@ -457,7 +517,17 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 		{
 			loaderSuccess = (this->*(format.loader))(file, loadFlags);
 			if(loaderSuccess)
+			{
+#if defined(MPT_BUILD_DEBUG)
+				// Verify that the probing function is consistent with our API contract
+				file.Rewind();
+				const auto data = file.GetRawDataAsByteVector(ProbeRecommendedSize);
+				MemoryFileReader mf{mpt::as_span(data)};
+				const uint64 size = file.GetLength();
+				MPT_ASSERT(format.prober(mf, &size) != ProbeFailure);
+#endif
 				break;
+			}
 		}
 
 		if(!loaderSuccess)
@@ -479,16 +549,58 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 	} else
 	{
 		// New song
-		InitializeGlobals();
+		InitializeGlobals(MOD_TYPE_NONE, 0);
 		m_visitedRows.Initialize(true);
 		m_dwCreatedWithVersion = Version::Current();
+#if MPT_TIME_UTC_ON_DISK
+#ifdef MODPLUG_TRACKER
+		if(GetType() & MOD_TYPE_IT)
+		{
+			m_modFormat.timezone = mpt::Date::LogicalTimezone::UTC;
+		} else
+		{
+			m_modFormat.timezone = mpt::Date::LogicalTimezone::Local;
+		}
+#else // !MODPLUG_TRACKER
+		if (GetType() & MOD_TYPE_IT)
+		{
+			m_modFormat.timezone = mpt::Date::LogicalTimezone::UTC;
+		} else
+		{
+			m_modFormat.timezone = mpt::Date::LogicalTimezone::Unspecified;
+		}
+#endif // MODPLUG_TRACKER
+#else
+#ifdef MODPLUG_TRACKER
+		m_modFormat.timezone = mpt::Date::LogicalTimezone::Local;
+#else // !MODPLUG_TRACKER
+		m_modFormat.timezone = mpt::Date::LogicalTimezone::Unspecified;
+#endif // MODPLUG_TRACKER
+#endif
 	}
 
+#if MPT_TIME_UTC_ON_DISK
+#ifdef MODPLUG_TRACKER
+	// convert timestamps to UTC
+	if(m_modFormat.timezone == mpt::Date::LogicalTimezone::Local)
+	{
+		for(auto & fileHistoryEntry : m_FileHistory)
+		{
+			if(fileHistoryEntry.HasValidDate())
+			{
+				fileHistoryEntry.loadDate = mpt::Date::forget_timezone(mpt::Date::UnixAsUTC(mpt::Date::UnixFromLocal(mpt::Date::interpret_as_timezone<mpt::Date::LogicalTimezone::Local>(fileHistoryEntry.loadDate))));
+			}
+		}
+		m_modFormat.timezone = mpt::Date::LogicalTimezone::UTC;
+	}
+#endif // MODPLUG_TRACKER
+#endif
+
 	// Adjust channels
 	const auto muteFlag = GetChannelMuteFlag();
-	for(CHANNELINDEX chn = 0; chn < MAX_BASECHANNELS; chn++)
+	for(CHANNELINDEX chn = 0; chn < ChnSettings.size(); chn++)
 	{
-		LimitMax(ChnSettings[chn].nVolume, uint16(64));
+		LimitMax(ChnSettings[chn].nVolume, uint8(64));
 		if(ChnSettings[chn].nPan > 256)
 			ChnSettings[chn].nPan = 128;
 		if(ChnSettings[chn].nMixPlugin > MAX_MIXPLUGINS)
@@ -504,15 +616,15 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 		sample.SanitizeLoops();
 
 #ifdef MPT_EXTERNAL_SAMPLES
-		if(SampleHasPath(nSmp))
+		if(SampleHasPath(nSmp) && (loadFlags & loadSampleData))
 		{
 			mpt::PathString filename = GetSamplePath(nSmp);
 			if(file.GetOptionalFileName())
 			{
-				filename = filename.RelativePathToAbsolute(file.GetOptionalFileName()->GetPath());
+				filename = mpt::RelativePathToAbsolute(filename, file.GetOptionalFileName()->GetDirectoryWithDrive());
 			} else if(GetpModDoc() != nullptr)
 			{
-				filename = filename.RelativePathToAbsolute(GetpModDoc()->GetPathNameMpt().GetPath());
+				filename = mpt::RelativePathToAbsolute(filename, GetpModDoc()->GetPathNameMpt().GetDirectoryWithDrive());
 			}
 			filename = filename.Simplify();
 			if(!LoadExternalSample(nSmp, filename))
@@ -556,13 +668,6 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 	m_nInstruments = maxInstr;
 
 	// Set default play state values
-	if(!m_nDefaultTempo.GetInt())
-		m_nDefaultTempo.Set(125);
-	else
-		LimitMax(m_nDefaultTempo, TEMPO(uint16_max, 0));
-	if(!m_nDefaultSpeed)
-		m_nDefaultSpeed = 6;
-
 	if(!m_nDefaultRowsPerBeat && m_nTempoMode == TempoMode::Modern)
 		m_nDefaultRowsPerBeat = 1;
 	if(m_nDefaultRowsPerMeasure < m_nDefaultRowsPerBeat)
@@ -570,13 +675,13 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 	LimitMax(m_nDefaultRowsPerBeat, MAX_ROWS_PER_BEAT);
 	LimitMax(m_nDefaultRowsPerMeasure, MAX_ROWS_PER_BEAT);
 	LimitMax(m_nDefaultGlobalVolume, MAX_GLOBAL_VOLUME);
-	LimitMax(m_nSamplePreAmp, 2000u);
-	LimitMax(m_nVSTiVolume, 2000u);
+	LimitMax(m_nSamplePreAmp, MAX_PREAMP);
+	LimitMax(m_nVSTiVolume, MAX_PREAMP);
 	if(!m_tempoSwing.empty())
 		m_tempoSwing.resize(m_nDefaultRowsPerBeat);
 
-	m_PlayState.m_nMusicSpeed = m_nDefaultSpeed;
-	m_PlayState.m_nMusicTempo = m_nDefaultTempo;
+	m_PlayState.m_nMusicSpeed = Order().GetDefaultSpeed();
+	m_PlayState.m_nMusicTempo = Order().GetDefaultTempo();
 	m_PlayState.m_nCurrentRowsPerBeat = m_nDefaultRowsPerBeat;
 	m_PlayState.m_nCurrentRowsPerMeasure = m_nDefaultRowsPerMeasure;
 	m_PlayState.m_nGlobalVolume = static_cast<int32>(m_nDefaultGlobalVolume);
@@ -597,7 +702,7 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 	if(UseFinetuneAndTranspose())
 		m_playBehaviour.reset(kPeriodsAreHertz);
 
-	m_nMaxOrderPosition = 0;
+	m_restartOverridePos = m_maxOrderPosition = 0;
 
 	RecalculateSamplesPerTick();
 
@@ -605,9 +710,7 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 	{
 		order.Shrink();
 		if(order.GetRestartPos() >= order.size())
-		{
 			order.SetRestartPos(0);
-		}
 	}
 
 	if(GetType() == MOD_TYPE_NONE)
@@ -615,7 +718,7 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 		return false;
 	}
 
-	SetModSpecsPointer(m_pModSpecs, GetBestSaveFormat());
+	m_pModSpecs = &GetModSpecifications(GetBestSaveFormat());
 
 	// When reading a file made with an older version of MPT, it might be necessary to upgrade some settings automatically.
 	if(m_dwLastSavedWithVersion)
@@ -652,6 +755,15 @@ bool CSoundFile::CreateInternal(FileReader file, ModLoadingFlags loadFlags)
 				{
 					// Plugin was found
 					plugin.pMixPlugin->RestoreAllParameters(plugin.defaultProgram);
+
+					// Special handling for instrument plugins in ProcessMixOps was removed
+					if(m_dwLastSavedWithVersion < MPT_V("1.32.00.11"))
+					{
+						if(plugin.pMixPlugin->IsInstrument())
+							plugin.SetMixMode(PluginMixMode::Instrument);
+						if(!plugin.pMixPlugin->GetNumInputChannels())
+							plugin.SetExpandedMix(false);
+					}
 				} else
 				{
 					// Plugin not found - add to list
@@ -723,6 +835,7 @@ bool CSoundFile::Destroy()
 	m_songArtist.clear();
 	m_songMessage.clear();
 	m_FileHistory.clear();
+	ChnSettings.clear();
 #ifdef MPT_EXTERNAL_SAMPLES
 	m_samplePaths.clear();
 #endif // MPT_EXTERNAL_SAMPLES
@@ -745,7 +858,7 @@ bool CSoundFile::Destroy()
 
 	m_nType = MOD_TYPE_NONE;
 	m_ContainerType = ModContainerType::None;
-	m_nChannels = m_nSamples = m_nInstruments = 0;
+	m_nSamples = m_nInstruments = 0;
 	return true;
 }
 
@@ -800,15 +913,15 @@ double CSoundFile::GetCurrentBPM() const
 void CSoundFile::ResetPlayPos()
 {
 	const auto muteFlag = GetChannelMuteFlag();
-	for(CHANNELINDEX i = 0; i < MAX_CHANNELS; i++)
+	for(CHANNELINDEX i = 0; i < m_PlayState.Chn.size(); i++)
 		m_PlayState.Chn[i].Reset(ModChannel::resetSetPosFull, *this, i, muteFlag);
 
 	m_visitedRows.Initialize(true);
-	m_SongFlags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
+	m_PlayState.m_flags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
 
 	m_PlayState.m_nGlobalVolume = m_nDefaultGlobalVolume;
-	m_PlayState.m_nMusicSpeed = m_nDefaultSpeed;
-	m_PlayState.m_nMusicTempo = m_nDefaultTempo;
+	m_PlayState.m_nMusicSpeed = Order().GetDefaultSpeed();
+	m_PlayState.m_nMusicTempo = Order().GetDefaultTempo();
 
 	// Do not ramp global volume when starting playback
 	m_PlayState.ResetGlobalVolumeRamping();
@@ -821,6 +934,9 @@ void CSoundFile::ResetPlayPos()
 	m_PlayState.m_nFrameDelay = 0;
 	m_PlayState.m_nextPatStartRow = 0;
 	m_PlayState.m_lTotalSampleCount = 0;
+	m_PlayState.m_ppqPosFract = 0.0;
+	m_PlayState.m_ppqPosBeat = 0;
+	m_PlayState.m_globalScriptState.Initialize(*this);
 }
 
 
@@ -868,9 +984,10 @@ void CSoundFile::SetCurrentOrder(ORDERINDEX nOrder)
 		m_PlayState.m_nPatternDelay = 0;
 		m_PlayState.m_nFrameDelay = 0;
 		m_PlayState.m_nextPatStartRow = 0;
+		m_PlayState.m_globalScriptState.Initialize(*this);
 	}
 
-	m_SongFlags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
+	m_PlayState.m_flags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
 }
 
 void CSoundFile::SuspendPlugins()
@@ -905,6 +1022,21 @@ void CSoundFile::ResumePlugins()
 }
 
 
+void CSoundFile::UpdatePluginPositions()
+{
+#ifndef NO_PLUGINS
+	for(auto &plugin : m_MixPlugins)
+	{
+		IMixPlugin *pPlugin = plugin.pMixPlugin;
+		if(pPlugin != nullptr && !pPlugin->IsResumed())
+		{
+			pPlugin->PositionChanged();
+		}
+	}
+#endif  // NO_PLUGINS
+}
+
+
 void CSoundFile::StopAllVsti()
 {
 #ifndef NO_PLUGINS
@@ -942,39 +1074,41 @@ void CSoundFile::RecalculateGainForAllPlugs()
 
 void CSoundFile::ResetChannels()
 {
-	m_SongFlags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
+	m_PlayState.m_flags.reset(SONG_FADINGSONG | SONG_ENDREACHED);
 	m_PlayState.m_nBufferCount = 0;
-	for(auto &chn : m_PlayState.Chn)
+	for(CHANNELINDEX channel = 0; channel < m_PlayState.Chn.size(); channel++)
 	{
+		ModChannel &chn = m_PlayState.Chn[channel];
+		chn.dwFlags.set(CHN_NOTEFADE);
 		chn.nROfs = chn.nLOfs = 0;
 		chn.nLength = 0;
+		chn.nFadeOutVol = 0;
 		if(chn.dwFlags[CHN_ADLIB] && m_opl)
-		{
-			CHANNELINDEX c = static_cast<CHANNELINDEX>(std::distance(std::begin(m_PlayState.Chn), &chn));
-			m_opl->NoteCut(c);
-		}
+			m_opl->NoteCut(channel);
 	}
 }
 
 
 #ifdef MODPLUG_TRACKER
 
-void CSoundFile::PatternTranstionChnSolo(const CHANNELINDEX chnIndex)
+void CSoundFile::PatternTranstionChnSolo(const CHANNELINDEX first, const CHANNELINDEX last)
 {
-	if(chnIndex >= m_nChannels)
+	if(first >= GetNumChannels() || last < first)
 		return;
 
-	for(CHANNELINDEX i = 0; i < m_nChannels; i++)
+	for(CHANNELINDEX i = 0; i < GetNumChannels(); i++)
 	{
-		m_bChannelMuteTogglePending[i] = !ChnSettings[i].dwFlags[CHN_MUTE];
+		if(i >= first && i <= last)
+			m_bChannelMuteTogglePending[i] = ChnSettings[i].dwFlags[CHN_MUTE];
+		else
+			m_bChannelMuteTogglePending[i] = !ChnSettings[i].dwFlags[CHN_MUTE];
 	}
-	m_bChannelMuteTogglePending[chnIndex] = ChnSettings[chnIndex].dwFlags[CHN_MUTE];
 }
 
 
 void CSoundFile::PatternTransitionChnUnmuteAll()
 {
-	for(CHANNELINDEX i = 0; i < m_nChannels; i++)
+	for(CHANNELINDEX i = 0; i < GetNumChannels(); i++)
 	{
 		m_bChannelMuteTogglePending[i] = ChnSettings[i].dwFlags[CHN_MUTE];
 	}
@@ -987,7 +1121,7 @@ void CSoundFile::LoopPattern(PATTERNINDEX nPat, ROWINDEX nRow)
 {
 	if(!Patterns.IsValidPat(nPat))
 	{
-		m_SongFlags.reset(SONG_PATTERNLOOP);
+		m_PlayState.m_flags.reset(SONG_PATTERNLOOP);
 	} else
 	{
 		if(nRow >= Patterns[nPat].GetNumRows()) nRow = 0;
@@ -997,7 +1131,7 @@ void CSoundFile::LoopPattern(PATTERNINDEX nPat, ROWINDEX nRow)
 		m_PlayState.m_nPatternDelay = 0;
 		m_PlayState.m_nFrameDelay = 0;
 		m_PlayState.m_nextPatStartRow = 0;
-		m_SongFlags.set(SONG_PATTERNLOOP);
+		m_PlayState.m_flags.set(SONG_PATTERNLOOP);
 	}
 	m_PlayState.m_nBufferCount = 0;
 }
@@ -1014,7 +1148,7 @@ void CSoundFile::DontLoopPattern(PATTERNINDEX nPat, ROWINDEX nRow)
 	m_PlayState.m_nFrameDelay = 0;
 	m_PlayState.m_nBufferCount = 0;
 	m_PlayState.m_nextPatStartRow = 0;
-	m_SongFlags.reset(SONG_PATTERNLOOP);
+	m_PlayState.m_flags.reset(SONG_PATTERNLOOP);
 }
 
 
@@ -1030,6 +1164,10 @@ PlayBehaviourSet CSoundFile::GetSupportedPlaybackBehaviour(MODTYPE type)
 	switch(type)
 	{
 	case MOD_TYPE_MPT:
+		playBehaviour.set(kOPLFlexibleNoteOff);
+		playBehaviour.set(kOPLwithNNA);
+		playBehaviour.set(kOPLNoteOffOnNoteChange);
+		[[fallthrough]];
 	case MOD_TYPE_IT:
 		playBehaviour.set(MSF_COMPATIBLE_PLAY);
 		playBehaviour.set(kPeriodsAreHertz);
@@ -1083,12 +1221,14 @@ PlayBehaviourSet CSoundFile::GetSupportedPlaybackBehaviour(MODTYPE type)
 		playBehaviour.set(kITDCTBehaviour);
 		playBehaviour.set(kITPitchPanSeparation);
 		playBehaviour.set(kITResetFilterOnPortaSmpChange);
-		if(type == MOD_TYPE_MPT)
-		{
-			playBehaviour.set(kOPLFlexibleNoteOff);
-			playBehaviour.set(kOPLwithNNA);
-			playBehaviour.set(kOPLNoteOffOnNoteChange);
-		}
+		playBehaviour.set(kITInitialNoteMemory);
+		playBehaviour.set(kITNoSustainOnPortamento);
+		playBehaviour.set(kITEmptyNoteMapSlotIgnoreCell);
+		playBehaviour.set(kITOffsetWithInstrNumber);
+		playBehaviour.set(kITDoublePortamentoSlides);
+		playBehaviour.set(kITCarryAfterNoteOff);
+		playBehaviour.set(kITNoteCutWithPorta);
+		playBehaviour.set(kITVolColNoSlidePropagation);
 		break;
 
 	case MOD_TYPE_XM:
@@ -1132,6 +1272,8 @@ PlayBehaviourSet CSoundFile::GetSupportedPlaybackBehaviour(MODTYPE type)
 		playBehaviour.set(kFT2PanSustainRelease);
 		playBehaviour.set(kFT2NoteDelayWithoutInstr);
 		playBehaviour.set(kFT2PortaResetDirection);
+		playBehaviour.set(kFT2AutoVibratoAbortSweep);
+		playBehaviour.set(kFT2OffsetMemoryRequiresNote);
 		break;
 
 	case MOD_TYPE_S3M:
@@ -1151,6 +1293,8 @@ PlayBehaviourSet CSoundFile::GetSupportedPlaybackBehaviour(MODTYPE type)
 		playBehaviour.set(kST3SampleSwap);
 		playBehaviour.set(kOPLNoteOffOnNoteChange);
 		playBehaviour.set(kApplyUpperPeriodLimit);
+		playBehaviour.set(kST3TonePortaWithAdlibNote);
+		playBehaviour.set(kS3MIgnoreCombinedFineSlides);
 		break;
 
 	case MOD_TYPE_MOD:
@@ -1219,6 +1363,8 @@ PlayBehaviourSet CSoundFile::GetDefaultPlaybackBehaviour(MODTYPE type)
 		// Default behaviour was chosen to follow GUS, so kST3PortaSampleChange is enabled and kST3SampleSwap is disabled.
 		// For SoundBlaster behaviour, those two flags would need to be swapped.
 		playBehaviour.reset(kST3SampleSwap);
+		// Most trackers supporting the S3M format, including all OpenMPT versions up to now, support fine slides with Kxy / Lxy, so only enable this quirk for files made with ST3.
+		playBehaviour.reset(kS3MIgnoreCombinedFineSlides);
 		break;
 
 	case MOD_TYPE_XM:
@@ -1266,6 +1412,8 @@ MODTYPE CSoundFile::GetBestSaveFormat() const
 		}
 		return MOD_TYPE_XM;
 	case MOD_TYPE_PSM:
+		if(Order.GetNumSequences() > 1)
+			return MOD_TYPE_MPT;
 		if(GetNumChannels() > 16)
 			return MOD_TYPE_IT;
 		for(CHANNELINDEX i = 0; i < GetNumChannels(); i++)
@@ -1326,29 +1474,6 @@ const char *CSoundFile::GetInstrumentName(INSTRUMENTINDEX nInstr) const
 }
 
 
-bool CSoundFile::InitChannel(CHANNELINDEX nChn)
-{
-	if(nChn >= MAX_BASECHANNELS)
-		return true;
-
-	ChnSettings[nChn].Reset();
-	m_PlayState.Chn[nChn].Reset(ModChannel::resetTotal, *this, nChn, GetChannelMuteFlag());
-
-#ifdef MODPLUG_TRACKER
-	if(GetpModDoc() != nullptr)
-	{
-		GetpModDoc()->SetChannelRecordGroup(nChn, RecordGroup::NoGroup);
-	}
-#endif // MODPLUG_TRACKER
-
-#ifdef MODPLUG_TRACKER
-	m_bChannelMuteTogglePending[nChn] = false;
-#endif // MODPLUG_TRACKER
-
-	return false;
-}
-
-
 void CSoundFile::InitAmigaResampler()
 {
 	if(m_SongFlags[SONG_ISAMIGA] && m_Resampler.m_Settings.emulateAmiga != Resampling::AmigaFilter::Off)
@@ -1516,7 +1641,7 @@ std::unique_ptr<CTuning> CSoundFile::CreateTuning12TET(const mpt::ustring &name)
 }
 
 
-mpt::ustring CSoundFile::GetNoteName(const ModCommand::NOTE note, const INSTRUMENTINDEX inst) const
+mpt::ustring CSoundFile::GetNoteName(const ModCommand::NOTE note, const INSTRUMENTINDEX inst, const NoteName *noteNames) const
 {
 	// For MPTM instruments with custom tuning, find the appropriate note name. Else, use default note names.
 	if(ModCommand::IsNote(note) && GetType() == MOD_TYPE_MPT && inst >= 1 && inst <= GetNumInstruments() && Instruments[inst] && Instruments[inst]->pTuning)
@@ -1524,7 +1649,7 @@ mpt::ustring CSoundFile::GetNoteName(const ModCommand::NOTE note, const INSTRUME
 		return Instruments[inst]->pTuning->GetNoteName(note - NOTE_MIDDLEC);
 	} else
 	{
-		return GetNoteName(note);
+		return GetNoteName(note, noteNames ? noteNames : m_NoteNames);
 	}
 }
 
@@ -1546,9 +1671,10 @@ mpt::ustring CSoundFile::GetNoteName(const ModCommand::NOTE note, const NoteName
 		return specialNoteNames[note - NOTE_MIN_SPECIAL];
 	} else if(ModCommand::IsNote(note))
 	{
+		const int octave = (note - NOTE_MIN) / 12;
 		return mpt::ustring()
 			.append(noteNames[(note - NOTE_MIN) % 12])
-			.append(1, UC_('0') + (note - NOTE_MIN) / 12)
+			.append(1, static_cast<mpt::uchar>((octave <= 9 ? UC_('0') : UC_('A') - 10) + octave))
 			;	// e.g. "C#" + "5"
 	} else if(note == NOTE_NONE)
 	{
@@ -1573,29 +1699,29 @@ const NoteName *CSoundFile::GetDefaultNoteNames()
 #endif // MODPLUG_TRACKER
 
 
-void CSoundFile::SetModSpecsPointer(const CModSpecifications*& pModSpecs, const MODTYPE type)
+const CModSpecifications &CSoundFile::GetModSpecifications(const MODTYPE type)
 {
 	switch(type)
 	{
 		case MOD_TYPE_MPT:
-			pModSpecs = &ModSpecs::mptm;
+			return ModSpecs::mptm;
 		break;
 
 		case MOD_TYPE_IT:
-			pModSpecs = &ModSpecs::itEx;
+			return ModSpecs::itEx;
 		break;
 
 		case MOD_TYPE_XM:
-			pModSpecs = &ModSpecs::xmEx;
+			return ModSpecs::xmEx;
 		break;
 
 		case MOD_TYPE_S3M:
-			pModSpecs = &ModSpecs::s3mEx;
+			return ModSpecs::s3mEx;
 		break;
 
 		case MOD_TYPE_MOD:
 		default:
-			pModSpecs = &ModSpecs::mod;
+			return ModSpecs::mod;
 			break;
 	}
 }
@@ -1605,7 +1731,7 @@ void CSoundFile::SetType(MODTYPE type)
 {
 	m_nType = type;
 	m_playBehaviour = GetDefaultPlaybackBehaviour(GetBestSaveFormat());
-	SetModSpecsPointer(m_pModSpecs, GetBestSaveFormat());
+	m_pModSpecs = &GetModSpecifications(GetBestSaveFormat());
 }
 
 
@@ -1615,7 +1741,7 @@ void CSoundFile::ChangeModTypeTo(const MODTYPE newType, bool adjust)
 {
 	const MODTYPE oldType = GetType();
 	m_nType = newType;
-	SetModSpecsPointer(m_pModSpecs, m_nType);
+	m_pModSpecs = &GetModSpecifications(m_nType);
 
 	if(oldType == newType || !adjust)
 		return;
@@ -1640,7 +1766,7 @@ void CSoundFile::ChangeModTypeTo(const MODTYPE newType, bool adjust)
 	Order.OnModTypeChanged(oldType);
 	Patterns.OnModTypeChanged(oldType);
 
-	m_modFormat.type = mpt::ToUnicode(mpt::Charset::UTF8, GetModSpecifications().fileExtension);
+	m_modFormat.type = GetModSpecifications().GetFileExtension();
 }
 
 #endif // MODPLUG_TRACKER
@@ -1702,7 +1828,7 @@ std::vector<SubSong> CSoundFile::GetAllSubSongs()
 		subSongs.reserve(subSongs.size() + subSongsSeq.size());
 		for(const auto &song : subSongsSeq)
 		{
-			subSongs.push_back({song.duration, song.startRow, song.endRow, song.lastRow, song.startOrder, song.endOrder, song.lastOrder, seq});
+			subSongs.push_back({song.duration, song.startRow, song.endRow, song.restartRow, song.startOrder, song.endOrder, song.restartOrder, seq});
 		}
 	}
 	return subSongs;
@@ -1757,7 +1883,7 @@ uint32 CSoundFile::GetTickDuration(PlayState &playState) const
 
 	case TempoMode::Modern:
 		{
-			double accurateBufferCount = static_cast<double>(m_MixerSettings.gdwMixingFreq) * (60.0 / (playState.m_nMusicTempo.ToDouble() * Util::mul32to64_unsigned(playState.m_nMusicSpeed, playState.m_nCurrentRowsPerBeat)));
+			double accurateBufferCount = static_cast<double>(m_MixerSettings.gdwMixingFreq) * (60.0 / (playState.m_nMusicTempo.ToDouble() * static_cast<double>(Util::mul32to64_unsigned(playState.m_nMusicSpeed, playState.m_nCurrentRowsPerBeat))));
 			const TempoSwing &swing = (Patterns.IsValidPat(playState.m_nPattern) && Patterns[playState.m_nPattern].HasTempoSwing())
 				? Patterns[playState.m_nPattern].GetTempoSwing()
 				: m_tempoSwing;
@@ -1805,10 +1931,8 @@ double CSoundFile::GetRowDuration(TEMPO tempo, uint32 speed) const
 		return static_cast<double>(2500 * speed) / tempo.ToDouble();
 
 	case TempoMode::Modern:
-		{
-			// If there are any row delay effects, the row length factor compensates for those.
-			return 60000.0 / tempo.ToDouble() / static_cast<double>(m_PlayState.m_nCurrentRowsPerBeat);
-		}
+		// If there are any row delay effects, the row length factor compensates for those.
+		return 60000.0 / tempo.ToDouble() / static_cast<double>(m_PlayState.m_nCurrentRowsPerBeat);
 
 	case TempoMode::Alternative:
 		return static_cast<double>(1000 * speed) / tempo.ToDouble();
@@ -1816,18 +1940,10 @@ double CSoundFile::GetRowDuration(TEMPO tempo, uint32 speed) const
 }
 
 
-const CModSpecifications& CSoundFile::GetModSpecifications(const MODTYPE type)
-{
-	const CModSpecifications* p = nullptr;
-	SetModSpecsPointer(p, type);
-	return *p;
-}
-
-
 ChannelFlags CSoundFile::GetChannelMuteFlag()
 {
 #ifdef MODPLUG_TRACKER
-	return (TrackerSettings::Instance().m_dwPatternSetup & PATTERN_SYNCMUTE) ? CHN_SYNCMUTE : CHN_MUTE;
+	return (TrackerSettings::Instance().patternSetup & PatternSetup::SyncMute) ? CHN_SYNCMUTE : CHN_MUTE;
 #else
 	return CHN_SYNCMUTE;
 #endif
@@ -1927,7 +2043,7 @@ bool CSoundFile::IsSampleReferencedByInstrument(SAMPLEINDEX sample, INSTRUMENTIN
 	if(targetIns == nullptr)
 		return false;
 
-	return mpt::contains(mpt::as_span(targetIns->Keyboard).first(NOTE_MAX), sample);
+	return mpt::contains(mpt::as_span(targetIns->Keyboard).first(NOTE_MAX - NOTE_MIN + 1), sample);
 }
 
 
@@ -1971,7 +2087,7 @@ void CSoundFile::PrecomputeSampleLoops(bool updateChannels)
 bool CSoundFile::LoadExternalSample(SAMPLEINDEX smp, const mpt::PathString &filename)
 {
 	bool ok = false;
-	InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
+	mpt::IO::InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
 
 	if(f.IsValid())
 	{
@@ -2005,20 +2121,20 @@ bool CSoundFile::LoadExternalSample(SAMPLEINDEX smp, const mpt::PathString &file
 #endif // MPT_EXTERNAL_SAMPLES
 
 
-// Set up channel panning and volume suitable for MOD + similar files. If the current mod type is not MOD, bForceSetup has to be set to true.
-void CSoundFile::SetupMODPanning(bool bForceSetup)
+// Set up channel panning suitable for MOD + similar files. If the current mod type is not MOD, forceSetup has to be set to true for this function to take effect.
+void CSoundFile::SetupMODPanning(bool forceSetup)
 {
 	// Setup LRRL panning, max channel volume
-	if(!(GetType() & MOD_TYPE_MOD) && bForceSetup == false) return;
+	if(!(GetType() & MOD_TYPE_MOD) && !forceSetup)
+		return;
 
-	for(CHANNELINDEX nChn = 0; nChn < MAX_BASECHANNELS; nChn++)
+	for(CHANNELINDEX chn = 0; chn < GetNumChannels(); chn++)
 	{
-		ChnSettings[nChn].nVolume = 64;
-		ChnSettings[nChn].dwFlags.reset(CHN_SURROUND);
+		ChnSettings[chn].dwFlags.reset(CHN_SURROUND);
 		if(m_MixerSettings.MixerFlags & SNDMIX_MAXDEFAULTPAN)
-			ChnSettings[nChn].nPan = (((nChn & 3) == 1) || ((nChn & 3) == 2)) ? 256 : 0;
+			ChnSettings[chn].nPan = (((chn & 3) == 1) || ((chn & 3) == 2)) ? 256 : 0;
 		else
-			ChnSettings[nChn].nPan = (((nChn & 3) == 1) || ((nChn & 3) == 2)) ? 0xC0 : 0x40;
+			ChnSettings[chn].nPan = (((chn & 3) == 1) || ((chn & 3) == 2)) ? 0xC0 : 0x40;
 	}
 }
 
